package rmrs.persist;

/*
 * Copyright (c) 1999 The Regenstrief Institute.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Written by Gunther Schadow.
 *
 * $Id: ObjectRelationalBroker.java,v 1.1 1999/12/06 19:25:44 schadow Exp $
 */

import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.*;

/** A persistence broker used with PostgreSQL (and perhaps other SQL3
 * data bases,) where an object can be stored in one table access.
 *
 * @author Gunther Schadow
 * @version $Id: ObjectRelationalBroker.java,v 1.1 1999/12/06 19:25:44 schadow Exp $ 
 */
public class ObjectRelationalBroker 
    extends RelationalBroker
{
    static final String OID_SEQUENCE   = "ps_oidsq";

    public ObjectRelationalBroker(String url, String user, String password)
	throws PersistException
    {
	super(url, user, password);
    }

    public ObjectRelationalBroker(Connection con) 
	throws PersistException
    {
	super(con);
    }
    
    // not for public use but only called from PersistentObject.save !!!
    public OID save(PersistentObject o) 
	throws PersistException
    {
	
	Class theClass = o.getClass();
	RelClassMapping classMapping = 
	    (RelClassMapping)classMap.get(theClass);
	String theTable = classMapping.theTable;
	OID oid = o.getOID();
	
	SQLStatement sql = null;

	// first save the root stuff and grab a new OID by the way
	try {
	    if(o.isNew()) {
		// here we can directly insert into the species table
		// and the ORDBMS will distribute the stuff over the
		// inheritance chaing properly
		sql = new SQLInsert(theTable);

		Statement stmt = con.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT NEXTVAL('" +
						 OID_SEQUENCE + "');");
		if(rs.next()) {
		    o.po_oid = oid = new OID(rs.getLong(1) + 1);
		    
		    sql.addField(OID_COLUMN, oid);	    
		    sql.addField(RT_TABLE, theTable);
		    sql.addField(RT_CLASS, theClass.getName());
		    sql.addField(RT_FLAGS, new Integer(o.po_flags));
		    sql.addField(RT_LAST_SAVE, o.getLastSaveTime());  
		    sql.addField(RT_CHECK_OUT, o.getCheckOutTime());  
		}
		
		rs.close();
		stmt.close();

	    } else {
		// here we can directly insert into the species table
		// and the ORDBMS will distribute the stuff over the
		// inheritance chaing properly
		sql = new SQLUpdate(theTable);
		
		sql.addKey(OID_COLUMN, oid);	    
		sql.addField(RT_FLAGS, new Integer(o.po_flags));
		sql.addField(RT_LAST_SAVE, o.getLastSaveTime());  
	    }
	    	    
	    while(theClass != PersistentObject.class) {
		doSave(o, sql, "", classMapping);
		theClass = theClass.getSuperclass();
		classMapping = (RelClassMapping)classMap.get(theClass);
	    }

	    // only one sql command executed per object, should be a 
	    // lot more efficient!
	    sql.execute(con);

	    // if we can figure out to do the commit only after we have
	    // saved an entire graph, that would be awesome!
	    con.commit();

	    sql.close();

	    return o.getOID();

	} catch(SQLException x) {
	    x.printStackTrace();
	    throw new Error(x.getMessage());
	} catch(IllegalAccessException x) {
	    x.printStackTrace();
	    throw new Error(x.getMessage());
	}
    }

    // load is public -- as opposed to save -- since the object to
    // retrieve is not yet present as a Java object
    public PersistentObject load(Class classToConformTo, OID oid) 
	throws PersistException
    {

	// Overview of Strategy:
	// (0) try to find object id in cache, and if so, return right away!
	// (1) get the entire object from the root
	// (2) check whether the species' class conforms to the argument
	// (2.5) read the entire object
	// (3) created that object and set the management values 
	// (4) read the fields beginning with the species all the way
	//     up to the superclass right below the root
	// (5) notify the newly restored object about it's new life
	// (6) stamp new access time, set advisory locks, release any
	//     row locks into the root

	// Here goes:
	// (1) find the super class

	// (0) try to find object id in cache, and if so, return right away!

	PersistentObject o = (PersistentObject)objectCache.get(oid);
	if(o != null)
	    return o;

	// else get into the data base

	try {
	    // (1) get the species' table and class from the root
	    
	    PreparedStatement stmt = con.prepareStatement("SELECT " +
							  RT_TABLE + ", " +
							  RT_CLASS +
							  " FROM " + 
							  ROOT_TABLE + 
							  "* WHERE " + 
							  OID_COLUMN + "=?;");
	    oid.setAsSQLParameter(stmt, 1);
	    ResultSet rs = stmt.executeQuery();
	    
	    if(! rs.next())
		throw new Error("oid " + oid + " not found in " + ROOT_TABLE);
	    
	    String theTable = rs.getString(RT_TABLE);
	    String theClassName = rs.getString(RT_CLASS);
	    Class  theClass = Class.forName(theClassName);
	    RelClassMapping classMapping = 
		(RelClassMapping)classMap.get(theClass);

	    // (2) check whether the species' class conforms to the argument
	    if(! classToConformTo.isAssignableFrom(theClass))
		throw new Error("class " + theClassName 
				+ " does not conform to class "
				+ classToConformTo.getName());

	    // (2.5) now read the entire object
	    stmt.close();
	    stmt = con.prepareStatement("SELECT * FROM " + 
					theTable + 
					" WHERE " + 
					OID_COLUMN + "=?;");
	    oid.setAsSQLParameter(stmt, 1);
	    rs = stmt.executeQuery();
	    
	    if(! rs.next())
		throw new Error("oid " + oid + " not found in " + theTable);
	    
	    // (3) create that object and set management values	    
	    if(classMapping.makeMethod == null)
		o = (PersistentObject)classMapping.theClass.newInstance();
	    else
		o = (PersistentObject)classMapping.makeMethod
		    .invoke(null, null);

	    o.prepareRestoring(this, oid, rs.getInt(RT_FLAGS),
			       rs.getTimestamp(RT_CHECK_OUT),
			       rs.getTimestamp(RT_LAST_SAVE));
	    
	    // (4) read the fields beginning with the species all the way
	    //     up to the superclass
	    while(theClass != PersistentObject.class) {
		// pre-register the object so the subsequent loads can grab
		// it from the cache.
		register(o);

		doLoad(o, rs, "", classMapping);

		theClass = theClass.getSuperclass();
		classMapping = (RelClassMapping)classMap.get(theClass);
	    }

	    // (5) notify the newly restored object about it's new life
	    
	    o.notifyRestored();

	    // (6) stamp new access time, set advisory locks, release any
	    //     row locks
	    // FIXME: feature not yet implemented
	    
	    return o;
	    
	} catch(SQLException x) {
	    x.printStackTrace();
	    throw new Error(x.getMessage());
	} catch(IllegalAccessException x) {
	    x.printStackTrace();
	    throw new Error(x.getMessage());
	} catch(InstantiationException x) {
	    x.printStackTrace();
	    throw new Error(x.getMessage());
	} catch(InvocationTargetException x) {
	    x.printStackTrace();
	    throw new Error(x.getMessage());
	} catch(ClassNotFoundException x) {
	    x.printStackTrace();
	    throw new Error(x.getMessage());
	}
    }


    public ObjectSet query(PersistentObject queryExample) 
	throws PersistException
    {
	// Strategy
	// 1. make a new SQLSelect statement
	// 2. add field ps_oid -- we go in two passes: first get the oid
	//    then load the data.
	// 3. cycle through the java inheritance chain
	// 3.1 direct mapping and non-null values:
	//       add criteria (name, value, [op])
	// 3.2 inline mapping: recurse (just like doSave)
	// 3.3 ref-to-one mapping and non-null value:
	// 3.3.1 if value is a real object: include oid as criteria
	// 3.3.2 if value is a query example: ... boy that gets hard!
	// 3.4 ref-to-many mapping and non-null value:
	// 3.4.1 if value is a real object: include oid as criteria
	// ... but how???

	return new ObjectSet(queryExample.getClass());
    }
}
