package rmrs.persist;

/*
 * Copyright (c) 1999 The Regenstrief Institute.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Written by Gunther Schadow.
 *
 * $Id: PersistException.java,v 1.1 1999/12/06 19:25:44 schadow Exp $
 */

/** We only have a single exception class that can represent several
 * exceptions. This is just more convenient that cluttering up the namespace
 * with different exceoption classes.
 *
 * @author Gunther Schadow
 * @version $Id: PersistException.java,v 1.1 1999/12/06 19:25:44 schadow Exp $ 
 * */
public class PersistException
  extends Exception
{
    /** A catch-all for unclassified or unpredictable exceptions. */
    public static final int MISC_EXCEPTION          = 0;

    /** Thrown if save, load or query is tried without a broker. */
    public static final int NO_BROKER               = 1;

    /** Thrown if one tries to reassign a broker to a real object that
     * already has a broker assigned. */
    public static final int OTHER_BROKER            = 2;

    /** Save was tried on an object that is write protected. An object
     * is write protected if it is not a real object or if a real object
     * was checked out as write protected.
     */
    public static final int WRITE_PROTECTED	    = 3;

    /** Exceptions from the underlying JDBC work. */
    public static final int SQL_EXCEPTION	    = 4;

    /** If there is no useful constructor that can be called by the
     * broker when retrieving an object. */
    public static final int INVALID_CONSTRUCTOR	    = 5;

    /** If a field was designated for inline storage but contains
     * a persistent object. Persistent object may never be stored
     * inline. Only composite values without identity can be inlined. 
     */
    public static final int NO_INLINE	            = 6;

    /** Class is not instantiable for strange reasons. */
    public static final int CLASS_NOT_INSTANTIABLE  = 7;

    /** Thrown if we can't access a field of a persistent object.
     * Should not normally occur, unless someone has constrained
     * some security policy?
     */
    public static final int NO_ACCESS_TO_FIELD	    = 8;

    /** Thrown if a java class can not be found. Should not 
     * occur dynamically but only when reading the initialization
     * tables */
    public static final int CLASS_NOT_FOUND	    = 9;

    /** The initialization tables mention a field that apparently
     * doesn't exist (any more?) */
    public static final int FIELD_NOT_FOUND	    = 10;

    /** This is an internal error that should never occur: the
     * same persistent object was instantiated twice in a JVM.
     */
    public static final int DUPLICATE_JAVA_OBJECT   = 11;

    /** The OID could not be found in the realm of the persistence
     * broker used to load an object */
    public static final int OID_NOT_FOUND	    = 12;

    /** The class found in the persistent store is not assignable
     * to the class requested. */
    public static final int CLASS_NOT_CONFORMANT    = 13;

    /** An exception was thrown when the object was initialized
     * for retrieval. */
    public static final int CONSTRUCTOR_EXCEPTION   = 14;

    /** A query example was expected, yet a real object (or object
	view) was provided. */
    public static final int QUERY_EXAMPLE_EXPECTED  = 15;

    public static final int OBJECT_VIEW_EXPECTED    = 16;
    public static final int REAL_OBJECT_EXPECTED    = 17;

    static final String texts[] = {
	"miscellaneous exception",
	"no persistence broker assigned to object",
	"persistence broker is already assigned to object",
	"object is write protected",
	"sql exception from lower layer",
	"invalid constructor",
	"persistent object must not be inlined",
	"java class is not instantiable (abstract or interface)",
	"can not get access to field",
	"java class not found",
	"java field not found",
	"duplicate java object for oid",
	"oid not found",
	"actual class does not conform to requested class",
	"constructor of loaded class has thrown exception",
	"query example expected",
	"object view expected",
	"real object expected"
    };
    
    int     type = MISC_EXCEPTION;
    Throwable ex = null;
    String   arg = null;
    
    public int       getType()      { return type; }
    public String    getArg()       { return arg; }
    public Throwable getException() { return ex; }
    
    /**
     * Raise a PersistException of specified type.
     */ 
    public PersistException(int type)
    { 
	super(texts[type]); 
	this.type = type;
    }
    
    /**
     * Raise a PersistException of specified type with a string argument.
     */ 
    public PersistException(int type, String arg)
    { 
	this(type, arg, null);
    }
    
    /**
     * This variant is called if we forward another exception
     * just caught.
     */
    public PersistException(int type, String arg, Throwable ex)
    { 
	super(texts[type] + ": " + arg + 
	      ( ex != null ? " : " + ex.getMessage() : "" )); 
	this.type = type;
	this.arg  = arg;
	this.ex   = ex;
    }
}
