package rmrs.persist;

/*
 * Copyright (c) 1999 The Regenstrief Institute.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Written by Gunther Schadow.
 *
 * $Id: PersistentObject.java,v 1.1 1999/12/06 19:25:44 schadow Exp $
 */

import java.util.Date;
import java.lang.reflect.Field;

/** Any class needing persistence for its object must inherit from
 * the class PersistentObject. To persist an object it needs a
 * persistence broker assigned to it. A peristence broker is assigned
 * to an object in either of two ways:
 * <OL><LI>when a <em>new</em> object is first saved to that broker
 * with the method <code>void save(PersistenceBroker broker)</code>,
 * <LI>when an object is retrieved from a persistence broker. To save
 * a previously retrieved object, just use the method <code>void save()
 * </code> without arguments.
 * </OL>
 * <P>
 * A persistence broker is statically assigned to each object. At
 * this point, an object can not migrate from one broker to another,
 * however, this could be made possible quite easily.
 * <P>
 * Note that persistence brokers are assigned to objects not to
 * entire classes. This means, that the objects of one class can
 * be distributed over multiple persistence brokers.
 * <P>
 * A class that inherits from PersistentObject only needs to
 * obey to a few conventions, as described in the following
 * checklist:
 * <OL>
 * <LI><P>
 * Provide either a public default constructor (i.e. a
 * constructor without arguments),
 * </P><LI><P>
 * or provide a static method without arguments that
 * returns a new java object of the given class. Using that latter
 * method is preferred since it can be private and thus protected
 * from accidential abuse outside the business of perstistence.
 * </P><LI><P>
 * To check the consistency of a retireved object and to reset
 * transient variables that were not stored, the persistent class
 * should provide a method <CODE>void notifyRestored()</CODE>. The
 * persistence broker will call this method when the object has
 * been retrieved.
 * </P></OL>
 * <P>
 * When an object is saved for the first time it will be assigned
 * a unique object id. Currently that object id is unique only
 * within the persistence broker, but (with only a few more work)
 * it can be made unique accross all objects.
 * <P>
 * A persistent object can be in any of three principle "moods".
 * This "mood" can not change throughout the life of a Java object.
 * <OL>
 * <LI><P>
 * A <B>real object</B> represents a persistent object in a Java VM.
 * Multiple JVMs can have the same persistent object checked out
 * at the same time, but each persistent object can only occur once
 * as a real object in any JVM. (Note: the concurrency is not
 * currently checked and handled reliably. Needs more work.)
 * </P><LI><P>
 * A <B>query example</B> is a java object of a persistent class
 * that is not in itself persistent, does not have an OID, and does
 * not represent an individuum. Conversely, a query example represents
 * an entire class (set) of objects that have are consistent with the
 * specified field values.
 * </P><LI><P>
 * An <B>object view</B> is a java object without itself being 
 * persistent reflecting other persistent objects one at a time.
 * An object view is like a mirror that can show different faces
 * without being itself a face. Object views will be used to
 * allow scanning through data base tables without incurring the
 * high price of Java object creation. Not currently used.
 * </P></OL>
 *
 * @author Gunther Schadow
 * @version $Id: PersistentObject.java,v 1.1 1999/12/06 19:25:44 schadow Exp $ 
 */
public class PersistentObject 
    implements Cloneable
{
    /** The oid of an object. Set as soon as the object is saved for
     * the first time. Only for real objects.
     */
    OID po_oid   = null;

    /** Management bits, used only through integer constants defined
     * next.
     */
    int po_flags = NEW | SAVEABLE;
    
    /** indicates the object is brand new, never saved before. */
    public static final int NEW           =               0x001;

    /** indicates that object is a real perisistent object,
     * it can be saved and retrieved.
     */
    public static final int REAL_OBJECT   =               0x002;

    /** indicates that object is a query example used in formulating
     * queries. The query method can be invoked only on a query
     * example.
     */
    public static final int QUERY_EXAMPLE =               0x004;

    /** indicates  that object is an object view, whose values
     * reflect the values of another real object.  Beware that
     * object views may change their values dynamically. Object
     * views are only used locally for scanning and searching
     * operations.
     */
    public static final int OBJECT_VIEW   =               0x008;

    /** object is a real object an not write protected. */
    public static final int SAVEABLE      = REAL_OBJECT | 0x010;

    /** object is supposed to be automatically saved upon
     * finalization.
     */
    public static final int AUTO_SAVE     = SAVEABLE    | 0x020;

    /** Checkout time allows the detection of concurrent 
     * check outs.  When an object is checked out the checkout
     * time in the database is set. When an object is saved
     * back and check out time has changed, we know that someone
     * else uses this object.  How is concurrency handled?
     * For example, by allowing only read-only check-outs on
     * an object that's already checked out.
     */
    Date po_checkOutTime = null;

    /** Last saved time allows to tell whether an object has 
     * changed since it was last saved. Concurrency strategy
     * needs more planning and implementation!
     */
    Date po_lastSaveTime = null;
    
    /** Each object has a persistence broker assigned to it,
     * on which it is saved, from which it was loaded, and to
     * which queries are issued (for query examples).
     */
    PersistenceBroker po_broker = null;

    /** A constructor called automatically for any new persisten
	object created. This is called for both, objects that are
	newly created by the applications and to be retrieved
	objects by the persistence broker. Therefore anything
	done here should make sense for both cases. 
    */
    protected PersistentObject() { }

    /** GET ACCESSORS ******/

    /** Get accessor to OID field. An oid is only assigned to
     * real objects after being saved the first time (loaded
     * objects obviously were saved before.) Object views have
     * the oid value, of the object they currently reflect,
     * but the object view itself has no identity.
     *
     * @return the oid of the object or null if no oid assigned.
     */
    public OID  getOID()          { return po_oid; }

    /** Get accessor to check out time. 
     * Checkout time allows the detection of concurrent 
     * check outs.  When an object is checked out the checkout
     * time in the database is set. When an object is saved
     * back and check out time has changed, we know that someone
     * else uses this object.  How is concurrency handled?
     * For example, by allowing only read-only check-outs on
     * an object that's already checked out.
     *
     * @return the check out time as a Date (or sql.Timestamp) object.
     */
    public Date getCheckOutTime() { return po_checkOutTime; }

    /** Get accessor to time the object was last saved time. 
     * Last saved time allows to tell whether an object has 
     * changed since it was last saved. Concurrency strategy
     * needs more planning and implementation!
     */
    public Date getLastSaveTime() { return po_lastSaveTime; }

    /** Get accessor to the management bitset. Interpret only
     * using the predefined masks, better not use it at all. */
    protected int  getFlags()     { return po_flags; } 

    /** Get currently assigned persistence broker. */
    public PersistenceBroker getPersistenceBroker() { 
	return po_broker;
    }

    /** PREDICATES ******/

    /** Predicate telling whether the object has never been saved. */
    public boolean isNew() { 
	return (po_flags & NEW) == NEW; 
    }

    /** Predicate telling whether the object can be saved. */
    public boolean isSaveable() { 
	return (po_flags & SAVEABLE) == SAVEABLE;
    }

    /** Predicate telling whether the object is supposed to be
     * saved on finalization.
     */
    public boolean isAutoSave() { 
	return (po_flags & AUTO_SAVE) == AUTO_SAVE;
    }
    
    /** Predicate telling whether the java object is a real object. */
    public boolean isRealObject() { 
	return (po_flags & REAL_OBJECT) == REAL_OBJECT;
    }

    /** Predicate telling whether the java object is a query example. */
    public boolean isQueryExample() { 
	return (po_flags & QUERY_EXAMPLE) == QUERY_EXAMPLE;
    }

    /** Predicate telling whether the java object is an object view. */
    public boolean isObjectView() { 
	return (po_flags & OBJECT_VIEW) == OBJECT_VIEW;
    }

    /** SET ACCESSORS ******/

    /** Automatically save object in finalization.
     * @param v true to turn on auto save, false to turn it off.
     */
    public void setAutoSave(boolean v) { 
	if(v)
	    po_flags |= AUTO_SAVE;
	else
	    po_flags &= ~AUTO_SAVE | SAVEABLE;
    }


    /** METHODS THAT DO THE JOB *******/

    /** Save an object with a persistence broker. This method can only
     * be used once for new objects that have never been saved before. The
     * persistence broker automatically assigns an OID to a registrant
     * and a registrant must not have a pre-assigned oid.  
     *
     * @param pb a persistence broker that can store this kind of
     * objects.
     *
     * @except PersistException - if any problems occured.
     */
    public void save(PersistenceBroker pb) 
	throws PersistException
    { 
	// FIXME: what if someone assigns another broker if one
	// existed before? This would migrate the object to a
	// new broker, which has ramifications!
	if(po_broker != null)
	    throw new PersistException(PersistException.OTHER_BROKER);
	
	this.po_broker = pb;
	save();
    }

    /** Save an object with a previously assigned persistence broker.
     * This method can only be used if the object has been saved before
     * or was previously retrieved from a persistence broker.
     *
     * @except PersistException - if any problems occured.
     */
    public void save() 
	throws PersistException    
    {
	if(po_broker != null)
	    if((po_flags & SAVEABLE) == SAVEABLE) {
		po_lastSaveTime = new Date();
		po_broker.save(this);
		po_flags &= ~NEW;
	    } else
		throw new PersistException(PersistException.WRITE_PROTECTED);
	else
	    throw new PersistException(PersistException.NO_BROKER);
    }
    

    /** Call-back method to check the consistency of a retireved
     * object and to reset transient variables that were not stored,
     * the persistent class should overload this method if it
     * maintains transient state. The persistence broker will call this
     * method when the object has been retrieved.
     * */
    protected void notifyRestored() {}

    /** The persistence broker calls this method shortly before retrieval
     * to set the management field. Few classes will want to overload this.
     * */
    protected void prepareRestoring(PersistenceBroker pb, OID oid, 
				    int flags, Date cotime, Date lstime) {
	this.po_oid            = oid;
	this.po_flags          = flags & ~NEW;
	this.po_broker         = pb;
	this.po_checkOutTime   = cotime;
	this.po_lastSaveTime   = lstime;
    }
    
    /** This is for the auto save functionality, but finalize is not
     * called at the end of the program, why? Is it not virtual?
     * */
    protected void finalize() 
	throws PersistException
    {
	if((po_flags & AUTO_SAVE) == AUTO_SAVE)
	    po_broker.save(this);
    }

    /** Cloning becomes important to build convenient query examples (QE)
       and object views (OV).  QEs and OVs are Java objects of the same
       class as the objects queried for, but they are not the real objects
       themselves. The QE object is passed to a Broker's query
       method and eventually defines the SQL 'WHERE' clause.  Results of
       a query are passed back in an OV. The OV is like a cursor and has
       a next method to skip to the next result record.

       <P>Cloning comes into place at three steps:
       <OL><LI><P>
       a user can clone a real object to create a QE object. The user
       can then change some values and set other values to null and pass
       the QE object to a persistence broker's query method.
       </P><LI><P>
       the persistence broker will clone a QE object to create an OV
       object, which it returns to present the query results.
       </P><LI><P>
       the user can clone an OV object to create the real object 
       currently represented by that OV.
       </P></OL><P>
       This query by example feature is very intuitive and saves the
       programmer from dealing with lots of artificial query-criteria
       objects. It is also independent of SQL, i.e., it could work as
       well in an OODB.
       <P>
       What is arguably nice is that the user deals with one application
       class (e.g., Employee) and can use it to create real objects, 
       QE objects and OV objects. This allows great uniformity since
       dealing with those objects is largely independent on their mood.
       (We call "mood" what distinguishes real objects from QE and QV.)
       <P>
       There is a slight risk, though, that may or may not turn out 
       to be crucial.  The programmer must never assign an OV object
       to other objects as if it where a real object, since an OV
       object changes its values farther down the road.  Same is true
       for QE objects, but since a QE object is passed to the data base
       and not returned from it, the mistake to use it right away will
       occur less likely.
       <P>
       So, all persistent objects are cloneable, but the clone method
       is not directly used. Instead, the methods cloneQueryExample,
       cloneObjectView, and cloneRealObject are used.
    */
    protected Object clone() {
	try {
	    return super.clone();
	} catch(CloneNotSupportedException x) { 
	    x.printStackTrace();
	    throw new Error("FATAL ERROR: clone must be supported!");
	}
    }

    /** Create a query example to query for all objects that
     * are like this object. After cloning one would typically
     * set some of the query example's 
     * fields to null in order to broaden the search. The
     * clone is shallow, that is, assiciated objects are 
     * selected by OID. In the future, we'll allow complex
     * query examples that let you retrieve object graph
     * patterns.
     * */
    public PersistentObject cloneQueryExample() 
    {
	PersistentObject qe = (PersistentObject)clone();

	// a query example is no longer a real object ...
	qe.po_flags = qe.po_flags | QUERY_EXAMPLE & 
	    ~(REAL_OBJECT | SAVEABLE | AUTO_SAVE);

	// ... and has no object identity
	qe.po_oid = null;

	// but it keeps it's persistence broker, on which it
	// will execute the query.

	return qe;
    }

    /** Query for all objects that are like this query example.
     * Only allowed for query examples (not for real objects or
     * object views.)  The persistence broker can be reassigned
     * to search matching objects on multiple brokers. This 
     * reassigns the persistence broker of the query example.
     *
     * @param pb a persistence broker
     *
     * @exept PersistException if something went wrong, partibularly
     *     if this object is not a query example.
     * */
    public ObjectSet query(PersistenceBroker pb) 
	throws PersistException
    {
	if((po_flags & QUERY_EXAMPLE) == QUERY_EXAMPLE) {
	    this.po_broker = pb;
	    return po_broker.query(this);
	} else
	    throw new 
		PersistException(PersistException.QUERY_EXAMPLE_EXPECTED,
				 " got " + Integer.toHexString(po_flags));
    }

    /** Query for all objects that are like this query example.
     * Only allowed for query examples (not for real objects or
     * object views.)
     *
     * @exept PersistException if something went wrong, partibularly
     *     if this object is not a query example.
     * */
    public ObjectSet query() 
	throws PersistException
    {
	if((po_flags & QUERY_EXAMPLE) == QUERY_EXAMPLE)
	    return po_broker.query(this);
	else
	    throw new 
		PersistException(PersistException.QUERY_EXAMPLE_EXPECTED,
				 " got " + Integer.toHexString(po_flags));
    }
}
