package rmrs.persist;

/*
 * Copyright (c) 1999 The Regenstrief Institute.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Written by Gunther Schadow.
 *
 * $Id: RelationalBroker.java,v 1.1 1999/12/06 19:25:44 schadow Exp $
 */

import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.*;

/** A persistence broker for relational data bases. Will use one table
 * per class along the path of generalization. All will use the OID as
 * a shared key.
 *
 * @author Gunther Schadow
 * @version $Id: RelationalBroker.java,v 1.1 1999/12/06 19:25:44 schadow Exp $ 
 */
public class RelationalBroker 
    implements PersistenceBroker
{
    protected Hashtable classMap;
    protected String url;
    protected String user;
    protected String password;
    protected Connection con;
    
    // FIXME: needs to be encapsulated to allow soft references
    // on Java 1.2
    protected Hashtable objectCache = new Hashtable();

    static final String ROOT_TABLE     = "ps_object";
    static final String OID_COLUMN     = "ps_oid";
    static final String RT_CLASS       = "ps_class";
    static final String RT_TABLE       = "ps_table";
    static final String RT_FLAGS       = "ps_flags";
    static final String RT_CHECK_OUT   = "ps_cotime";
    static final String RT_LAST_SAVE   = "ps_lstime";
    static final String CLASS_TABLE    = "ps_class";
    static final String FIELD_TABLE    = "ps_field";

    public RelationalBroker(String url, String user, String password) 
	throws PersistException
    {
	try {
	    this.con = DriverManager.getConnection(url, user, password);
	    initializeClassMap();
	} catch (SQLException x) {
	    throw new PersistException(PersistException.SQL_EXCEPTION, 
				       "connect to " + url + 
				       " as " + user, x);
	}
    }

    public RelationalBroker(Connection con) 
	throws PersistException
    {
	this.con = con;
	initializeClassMap();
    }
    
    protected void initializeClassMap() 
	throws PersistException
    {
	classMap = new Hashtable();

	try {
	    Statement stmt = con.createStatement();
	    ResultSet rs = stmt.executeQuery("SELECT * FROM " + 
					     CLASS_TABLE + ";");
	    while(rs.next()) {
		RelClassMapping classMapping = new RelClassMapping(rs, con);
		classMap.put(classMapping.theClass, classMapping); 
	    }
	} catch(SQLException x) {
	    throw new PersistException(PersistException.SQL_EXCEPTION, 
				       "while initializing class map", x);
	}
    }
    
    protected void register(PersistentObject o) 
	throws PersistException
    {
	PersistentObject old = 
	    (PersistentObject)objectCache.put(o.getOID(), o);
	
	if(old != null && old != o)
	    throw new PersistException(PersistException.DUPLICATE_JAVA_OBJECT,
				       "oid " + o.getOID() + 
				       " for both " + old + " and " + o);
    }



    /** SAVE ********************************/

    // not for public use but only called from PersistentObject.save !!!
    public OID save(PersistentObject o) 
	throws PersistException
    {
	Class theClass = o.getClass();
	RelClassMapping classMapping = (RelClassMapping)classMap.get(theClass);
	String theTable = classMapping.theTable;
	OID oid = o.getOID();
	
	SQLStatement sql = null;

	// first save the root stuff and grab a new OID by the way
	try {
	    if(o.isNew()) {
		sql = new SQLInsert(ROOT_TABLE);

		// FIXME: this needs a lock on the OID index!!!
		// e.g. adding "FOR UPDATE" with certain systems
		// ASUGG: it can use a precompiled prepared statement!
		
		Statement stmt = con.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT MAX(" + 
						 OID_COLUMN + 
						 ") FROM " + 
						 ROOT_TABLE + ";");
		if(rs.next()) {
		    o.po_oid = oid = new OID(rs.getLong(1) + 1);
		    
		    sql.addField(OID_COLUMN, oid);	    
		    sql.addField(RT_TABLE, theTable);
		    sql.addField(RT_CLASS, theClass.getName());
		    sql.addField(RT_FLAGS, new Integer(o.po_flags));
		    sql.addField(RT_LAST_SAVE, o.getLastSaveTime());  
		    sql.addField(RT_CHECK_OUT, o.getCheckOutTime());  
		}
		
		sql.execute(con);
		rs.close();
		stmt.close();
		sql.close();

	    } else {
		
		sql = new SQLUpdate(ROOT_TABLE);
		
		sql.addKey(OID_COLUMN, oid);	    
		sql.addField(RT_FLAGS, new Integer(o.po_flags));
		sql.addField(RT_LAST_SAVE, o.getLastSaveTime());  

		sql.execute(con);
		sql.close();

	    }
	    	    
	    while(theClass != PersistentObject.class) {

		if(o.isNew())
		    sql = new SQLInsert(classMapping.theTable);
		else
		    sql = new SQLUpdate(classMapping.theTable);
		
		sql.addKey(OID_COLUMN, oid);

		doSave(o, sql, "", classMapping);

		sql.execute(con);

		theClass = theClass.getSuperclass();
		classMapping = (RelClassMapping)classMap.get(theClass);
		sql.close();

	    }

	    con.commit();
	    return o.getOID();

	} catch(SQLException x) {
	    throw new PersistException(PersistException.SQL_EXCEPTION, 
				       "in save", x);
	} catch(IllegalAccessException x) {
	    throw new PersistException(PersistException.NO_ACCESS_TO_FIELD, 
				       "in save", x);
	}
    }

    protected void doSave(Object o,
			  SQLStatement sql,
			  String prefix,
			  RelClassMapping classMapping) 
	throws PersistException, IllegalAccessException
    {
	
	Enumeration e = classMapping.fieldMap.elements();
	
	while(e.hasMoreElements()) {
	    RelFieldMapping fieldMapping = 
		(RelFieldMapping)e.nextElement();
	    
	    /* System.out.println("SAVE " + o.getClass().getName() + "." +
			       fieldMapping.theField.getName() + " -> " +
			       prefix + "-" + fieldMapping.theColumn + " " +
			       Integer.toHexString(fieldMapping.modifier));
	    */

	    if((fieldMapping.modifier & RelFieldMapping.DIRECT) 
	       == RelFieldMapping.DIRECT) {
		
		sql.addField(prefix + fieldMapping.theColumn, 
			     fieldMapping.theField.get(o));
		
	    } else if((fieldMapping.modifier & RelFieldMapping.INLINE)
		      == RelFieldMapping.INLINE) {
		
		// since we have tested it in initialization of this
		// broker, we know that an inlined object is never 
		// a persistent object
		
		Object inlinedObject = fieldMapping.theField.get(o);

		for(Class inlinedClass = inlinedObject.getClass();
		    inlinedClass != Object.class;
		    inlinedClass = inlinedClass.getSuperclass()) {
		    
		    // System.out.println(inlinedClass.getName());
		    
		    RelClassMapping inlinedClassMapping = 
			(RelClassMapping)classMap.get(inlinedClass);
		    
		    doSave(inlinedObject, 
			   sql, 
			   prefix + fieldMapping.theColumn, 
			   inlinedClassMapping);
		}
		
	    } else if((fieldMapping.modifier & RelFieldMapping.REF_TO_ONE) 
		      == RelFieldMapping.REF_TO_ONE) {
		
		PersistentObject associatedObject = 
		    (PersistentObject)fieldMapping.theField.get(o);
		
		if(associatedObject != null) {
		    if((fieldMapping.modifier & RelFieldMapping.CASCADE_SAVE)
		       == RelFieldMapping.CASCADE_SAVE)
			try {
			    associatedObject.save();
			} catch(PersistException x) {
			    if(x.getType() != PersistException.NO_BROKER)
				throw x;
			}

		    OID associatedOID = associatedObject.getOID();
		    if(associatedOID != null)
			sql.addField(prefix + fieldMapping.theColumn, 
				     associatedOID);		    
		}
	    } else if((fieldMapping.modifier & RelFieldMapping.REF_TO_MANY) 
		      == RelFieldMapping.REF_TO_MANY) {
		
		ObjectSet associatedSet = 
		    (ObjectSet)fieldMapping.theField.get(o);
		
		Enumeration ae = associatedSet.elements();
		while(ae.hasMoreElements()) {
		    PersistentObject associatedObject = 
			(PersistentObject)ae.nextElement();
		
		    if(associatedObject != null) {
			// need to assure that the proper backwards link
			// is set in every distal object 
			fieldMapping.distalField.set(associatedObject, o);

			// System.out.println(((PersistentObject)o).getOID());
			
			if((fieldMapping.modifier 
			    & RelFieldMapping.CASCADE_SAVE)
			   == RelFieldMapping.CASCADE_SAVE)
			    try {
				associatedObject.save();
			    } catch(PersistException x) {
				if(x.getType() != PersistException.NO_BROKER)
				    throw x;
			    }
		    }
		}
	    }
	}
    }



    /** LOAD ********************************/

    // load is public -- as opposed to save -- since the object to
    // retrieve is not yet present as a Java object
    public PersistentObject load(Class classToConformTo, OID oid) 
	throws PersistException
    {

	// Overview of Strategy:
	// (0) try to find object id in cache, and if so, return right away!
	// (1) get the species' table and class from the root
	// (2) check whether the species' class conforms to the argument
	// (3) created that object and set the management values 
	// (4) read the fields beginning with the species all the way
	//     up to the superclass right below the root
	// (5) notify the newly restored object about it's new life
	// (6) stamp new access time, set advisory locks, release any
	//     row locks into the root

	// Here goes:
	// (1) find the super class

	// (0) try to find object id in cache, and if so, return right away!

	PersistentObject o = (PersistentObject)objectCache.get(oid);
	if(o != null)
	    return o;

	// else get into the data base

	try {
	    // (1) get the species' table and class from the root
	    
	    PreparedStatement stmt = con.prepareStatement("SELECT * FROM " + 
							  ROOT_TABLE + 
							  " WHERE " + 
							  OID_COLUMN + "=?;");
	    oid.setAsSQLParameter(stmt, 1);
	    ResultSet rs = stmt.executeQuery();
	    
	    if(! rs.next())
		throw new PersistException(PersistException.OID_NOT_FOUND,
					   oid + " in " + ROOT_TABLE);
	    
	    String theTable = rs.getString(RT_TABLE);
	    String theClassName = rs.getString(RT_CLASS);
	    Class  theClass = Class.forName(theClassName);
	    RelClassMapping classMapping = 
		(RelClassMapping)classMap.get(theClass);

	    // (2) check whether the species' class conforms to the argument
	    if(! classToConformTo.isAssignableFrom(theClass))
		throw new PersistException(PersistException
					   .CLASS_NOT_CONFORMANT,
					   " actual: " + theClassName +
					   " requested: " 
					   + classToConformTo.getName());
	    
	    // (3) create that object and set management values	    
	    if(classMapping.makeMethod == null)
		o = (PersistentObject)classMapping.theClass.newInstance();
	    else
		o = (PersistentObject)classMapping.makeMethod
		    .invoke(null, null);

	    o.prepareRestoring(this, oid, rs.getInt(RT_FLAGS),
			       rs.getTimestamp(RT_CHECK_OUT),
			       rs.getTimestamp(RT_LAST_SAVE));
	    
	    // (4) read the fields beginning with the species all the way
	    //     up to the superclass
	    while(theClass != PersistentObject.class) {

		stmt = con.prepareStatement("SELECT * FROM " + 
					    classMapping.theTable + 
					    " WHERE " + OID_COLUMN + 
					    "=?;");
		oid.setAsSQLParameter(stmt, 1);
		rs = stmt.executeQuery();
		
		if(! rs.next())
		    throw new PersistException(PersistException.OID_NOT_FOUND,
				    "oid " + oid + "for class " + 
				    classMapping.theClass.getName() +
				    " in " + classMapping.theTable);
		
		register(o);
		doLoad(o, rs, "", classMapping);

		theClass = theClass.getSuperclass();
		classMapping = (RelClassMapping)classMap.get(theClass);
	    }

	    // (5) notify the newly restored object about it's new life
	    
	    o.notifyRestored();

	    // (6) stamp new access time, set advisory locks, release any
	    //     row locks
	    // FIXME: feature not yet implemented
	    
	    return o;
	    
	} catch(SQLException x) {
	    throw new PersistException(PersistException.SQL_EXCEPTION, 
				       "in load", x);
	} catch(IllegalAccessException x) {
	    throw new PersistException(PersistException.NO_ACCESS_TO_FIELD, 
				       "in save", x);
	} catch(InstantiationException x) {
	    throw new PersistException(PersistException.CLASS_NOT_INSTANTIABLE,
				       "in load", x);
	} catch(InvocationTargetException x) {
	    throw new PersistException(PersistException.CONSTRUCTOR_EXCEPTION,
				       "in load", x.getTargetException());
	} catch(ClassNotFoundException x) {
	    throw new PersistException(PersistException.CLASS_NOT_FOUND,
				       "in load", x);
	}
    }

    protected void doLoad(Object o,
			  ResultSet rs,
			  String prefix,
			  RelClassMapping classMapping) 
	throws PersistException, IllegalAccessException, 
	       SQLException, InstantiationException
    {	
	Enumeration e = classMapping.fieldMap.elements();
	
	while(e.hasMoreElements()) {
	    RelFieldMapping fieldMapping = 
		(RelFieldMapping)e.nextElement();
	    
	    /* System.out.println("LOAD " + o.getClass().getName() + "." +
			       fieldMapping.theField.getName() + " -> " +
			       prefix + "-" + fieldMapping.theColumn + " " +
			       Integer.toHexString(fieldMapping.modifier));
	    */

	    if((fieldMapping.modifier & RelFieldMapping.DIRECT) 
	       == RelFieldMapping.DIRECT) {
		
		fieldMapping.theField.set(o, 
		    rs.getObject(prefix + fieldMapping.theColumn));
		
	    } else if((fieldMapping.modifier & RelFieldMapping.INLINE)
		      == RelFieldMapping.INLINE) {
		
		// FIXME: inlined classes do not support subclass
		// polymorphism! I.e., you can only inline leaf-
		// level classes (final)!

		// The inlined object must have a public default
		// constructor

		Object inlinedObject =
		    fieldMapping.theField.getType().newInstance();
		
		for(Class inlinedClass = inlinedObject.getClass();
		    inlinedClass != Object.class;
		    inlinedClass = inlinedClass.getSuperclass()) {
		    
		    RelClassMapping inlinedClassMapping = 
			(RelClassMapping)classMap.get(inlinedClass);
		    
		    doLoad(inlinedObject, 
			   rs, 
			   prefix + fieldMapping.theColumn, 
			   inlinedClassMapping);		    
		}

		fieldMapping.theField.set(o, inlinedObject);
		
	    } else if((fieldMapping.modifier & RelFieldMapping.REF_TO_ONE) 
		      == RelFieldMapping.REF_TO_ONE) {
		
		// for this maneuver o must be a persistent object
		PersistentObject po = (PersistentObject)o;

		// FIXME: need to handle on-demand links (stubs?)
		
		OID associatedOID = 
		    OID.getFromSQLResult(rs, prefix + fieldMapping.theColumn);
		
		// pre-register the object so the subsequent loads can grab
		// it from the cache.
		register(po);

		if(associatedOID != null) {
		    PersistentObject associatedObject =
			load(fieldMapping.theField.getType(), 
			     associatedOID);
		    
		    fieldMapping.theField.set(o, associatedObject);
		} else 
		    fieldMapping.theField.set(o, null);

	    } else if((fieldMapping.modifier & RelFieldMapping.REF_TO_MANY) 
		      == RelFieldMapping.REF_TO_MANY) {
		
		// for this maneuver o must be a persistent object
		PersistentObject po = (PersistentObject)o;
		
		// class cast exception can be tested for in initialization!!
		// same with null pointer exception, every class must have
		// the set assigned before loading!
		ObjectSet associatedSet = 
		    (ObjectSet)fieldMapping.theField.get(po);

		associatedSet.registerLink(this, fieldMapping.distalField, po);
		
		RelClassMapping distalClassMapping = 
		    (RelClassMapping)classMap.get(fieldMapping.distalClass);

		PreparedStatement stmt = 
		    con.prepareStatement("SELECT " + OID_COLUMN + " FROM " + 
					 distalClassMapping.theTable +
					 " WHERE " + 
					 fieldMapping.theColumn + 
					 "=?;");
		po.getOID().setAsSQLParameter(stmt, 1);
		ResultSet ars = stmt.executeQuery();

		// pre-register the object so the subsequent loads can grab
		// it from the cache.
		register(po);

		while(ars.next()) {
		    OID associatedOID = 
			OID.getFromSQLResult(ars, OID_COLUMN);
		    
		    if((fieldMapping.modifier & RelFieldMapping.CASCADE_LOAD)
		       == RelFieldMapping.CASCADE_LOAD) {
			PersistentObject associatedObject = 
			    load(distalClassMapping.theClass, associatedOID);
			
			associatedSet.add(associatedObject);

			// set the backwards link in distal object 
			fieldMapping.distalField.set(associatedObject, po);
		    } else {
			associatedSet.add(associatedOID);
		    }
		}
	    }
	}
    }


    /** QUERY ********************************/

    public ObjectSet query(PersistentObject queryExample) 
	throws PersistException
    {
	// Strategy
	// 1. make a new SQLSelect statement
	// 2. add field ps_oid -- we go in two passes: first get the oid
	//    then load the data.
	// 3. cycle through the java inheritance chain
	// 3.1 direct mapping and non-null values:
	//       add criteria (name, value, [op])
	// 3.2 inline mapping: recurse (just like doSave)
	// 3.3 ref-to-one mapping and non-null value:
	// 3.3.1 if value is a real object: include oid as criteria
	// 3.3.2 if value is a query example: ... boy that gets hard!
	// 3.4 ref-to-many mapping and non-null value:
	// 3.4.1 if value is a real object: include oid as criteria
	// ... but how???
	
	try {
	    Class theClass = queryExample.getClass();
	    RelClassMapping classMapping = 
		(RelClassMapping)classMap.get(theClass);
	    String theTable = classMapping.theTable;
	    String oldTable = null;
	    
	    SQLSelect sql = new SQLSelect();
	    sql.addField(theTable + "." + OID_COLUMN);

	    while(theClass != PersistentObject.class) {
		theTable     = classMapping.theTable;

		sql.addTable(theTable);
		if(oldTable != null)
		    sql.addCriterion(theTable + "." + OID_COLUMN +
				     SQLSelect.op_eq +
				     oldTable + "." + OID_COLUMN);

		buildQuery(queryExample, sql, "", classMapping);
		
		oldTable     = theTable;
		theClass     = theClass.getSuperclass();
		classMapping = (RelClassMapping)classMap.get(theClass);
	    }

	    ResultSet rs = sql.executeQuery(con);

	    ObjectSet obs = new ObjectSet(queryExample.getClass());
	    obs.assignPersistenceBroker(this);
	    while(rs.next()) {
		OID oid = OID.getFromSQLResult(rs, 1);
		obs.add(oid);
	    }
	    return obs;
	    
	} catch(SQLException x) {
	    throw new PersistException(PersistException.SQL_EXCEPTION, 
				       "in save", x);
	} catch(IllegalAccessException x) {
	    throw new PersistException(PersistException.NO_ACCESS_TO_FIELD, 
				       "in save", x);
	}
    }
    

    protected void buildQuery(Object queryExample,
			      SQLSelect sql,
			      String prefix,
			      RelClassMapping classMapping) 
	throws PersistException, IllegalAccessException
    {
	Enumeration e = classMapping.fieldMap.elements();
	
	while(e.hasMoreElements()) {
	    RelFieldMapping fieldMapping = 
		(RelFieldMapping)e.nextElement();
	    
	    /* System.out.println("BUILD QUERY " + o.getClass().getName() + 
	                       "." +
			       fieldMapping.theField.getName() + " -> " +
			       prefix + "-" + fieldMapping.theColumn + " " +
			       Integer.toHexString(fieldMapping.modifier));
	    */

	    // disregard any field that is set to null 
	    if(fieldMapping.theField.get(queryExample) == null)
		continue;
		
	    if((fieldMapping.modifier & RelFieldMapping.DIRECT) 
	       == RelFieldMapping.DIRECT) {
		
		sql.addCriterion(prefix + fieldMapping.theColumn, 
				 SQLSelect.op_eq,
				 fieldMapping.theField.get(queryExample));
		
	    } else if((fieldMapping.modifier & RelFieldMapping.INLINE)
		      == RelFieldMapping.INLINE) {
		
		// since we have tested it in initialization of this
		// broker, we know that an inlined object is never 
		// a persistent object
		
		Object inlinedObject = fieldMapping.theField.get(queryExample);

		for(Class inlinedClass = inlinedObject.getClass();
		    inlinedClass != Object.class;
		    inlinedClass = inlinedClass.getSuperclass()) {
		    
		    // System.out.println(inlinedClass.getName());
		    
		    RelClassMapping inlinedClassMapping = 
			(RelClassMapping)classMap.get(inlinedClass);
		    
		    buildQuery(inlinedObject, 
			       sql, 
			       prefix + fieldMapping.theColumn, 
			       inlinedClassMapping);
		}
		
	    } else if((fieldMapping.modifier & RelFieldMapping.REF_TO_ONE) 
		      == RelFieldMapping.REF_TO_ONE) {
		
		PersistentObject associatedObject = 
		    (PersistentObject)fieldMapping.theField.get(queryExample);
		
		OID associatedOID = associatedObject.getOID();
		if(associatedOID != null)
		    sql.addCriterion(prefix + fieldMapping.theColumn, 
				     SQLSelect.op_eq, associatedOID);

	    } else if((fieldMapping.modifier & RelFieldMapping.REF_TO_MANY) 
		      == RelFieldMapping.REF_TO_MANY) {

		/* FIXME We don't really handle this here

		ObjectSet associatedSet = 
		    (ObjectSet)fieldMapping.theField.get(queryExample);
		
		Enumeration ae = associatedSet.elements();
		while(ae.hasMoreElements()) {
		    PersistentObject associatedObject = 
			(PersistentObject)ae.nextElement();
		
		    if(associatedObject != null) {
			fieldMapping.distalField.set(associatedObject,
						     queryExample);

			// System.out.println(((PersistentObject)o).getOID());
			
			if((fieldMapping.modifier 
			    & RelFieldMapping.CASCADE_SAVE)
			   == RelFieldMapping.CASCADE_SAVE)
			    try {
				associatedObject.save();
			    } catch(PersistException x) {
				if(x.getType() != PersistException.NO_BROKER)
				    throw x;
			    }
		    }
		}
		*/
	    }
	}
 
    }
}
